# encoding: utf-8

# included for ease of use with Python 2 (which has no statistics package)


from .utils import isnan, reject


def mean(values):
    values = list(reject(isnan, values))
    sigma = float(sum(values))
    n = len(values)
    if n:
        return sigma / n
    else:
        return None

def quantile(p):
    def bound_quantile(values):
        values = [value for value in sorted(reject(isnan, values))]
        n = len(values)
        ix = int(n * p)

        if not n:
            return None
        elif n % 2:
            return values[ix]
        elif ix < 1:
            return values[0]
        else:
            return mean(values[ix - 1:ix + 1])

    return bound_quantile

Q0 = min
Q1 = quantile(0.25)
Q2 = median = quantile(0.5)
Q3 = quantile(0.75)
Q4 = max



# encoding: utf-8


import builtins
import collections
import functools
import math


def isnan(value):
    return value is None or math.isnan(value)


def reject(predicate, values):
    return filter(lambda value: not predicate(value), values)


def replace(string, mapping):
    for match, replacement in mapping.items():
        string = string.replace(match, replacement)
    return string


def bound(value, lower, upper):
    return max(lower, min(upper, value))


def split(number):
    whole = int(number)
    fraction = number - whole
    return (whole, fraction)


def invert(number, default=None):
    if number:
        return 1 / number
    else:
        if default is None:
            return math.copysign(float('inf'), number)
        else:
            return default


def quantize(number, digits=0, q=builtins.round):
    """
    Quantize to somewhere in between a magnitude.

    For example:

        * ceil(55.25, 1.2) => 55.26
        * floor(55.25, 1.2) => 55.24
        * round(55.3333, 2.5) => 55.335
        * round(12.345, 1.1) == round(12.345, 2) == 12.34

    """
    base, fraction = split(digits)

    # quantization beyond an order of magnitude results in a variable amount
    # of decimal digits depending on the lowest common multiple,
    # e.g. floor(1.2341234, 1.25) = 1.225 but floor(1.2341234, 1.5) = 1.20
    if fraction * 10 % 1 > 0:
        digits = base + 2
    else:
        digits = base + 1

    multiplier = 10 ** base * invert(fraction, default=1)
    quantized = q(number * multiplier) / multiplier

    # additional rounding step to get rid of floating point math wonkiness
    return builtins.round(quantized, digits)


floor = functools.partial(quantize, q=math.floor)
ceil = functools.partial(quantize, q=math.ceil)
round = quantize


def repel(value):
    return math.copysign(max(abs(value), 1e-24), value)


def unwrap(fn):
    @functools.wraps(fn)
    def unwrapped_function(values, *vargs, **kwargs):
        scalar = not isinstance(values, collections.abc.Iterable)

        if scalar:
            values = [values]

        results = fn(values, *vargs, **kwargs)

        if scalar:
            results = results[0]

        return results

    return unwrapped_function


def vectorize(fn):
    """
    Allows a method to accept a list argument, but internally deal only
    with a single item of that list.
    """

    @functools.wraps(fn)
    def vectorized_function(values, *vargs, **kwargs):
        return [fn(value, *vargs, **kwargs) for value in values]

    return vectorized_function



# encoding: utf-8

import itertools
import unittest

import pandas

from . import notation

cases = [
    ([11234.22, 233000.55, 1175125.2], ['11K', '233K', '1,180K']),
    ([111, 1111.23, 1175125.234], ['0.11K', '1.11K', '1,180.00K']),
    ([10 ** 16, 10 ** 15, 10 ** 14], ['10,000T', '1,000T', '100T']),
    ([10 ** -3, 10 ** -6, 10 ** -9], ['1,000.00µ', '1.00µ', '0.00µ']),
    ([10 ** -16, 10 ** -15, 10 ** -14], ['0.00010p', '0.00100p', '0.01000p']),
    ]

missing_cases = [
    ([None], ['']),
    ([None, 5, 2, 1, 1, 1, 2, None], ['', '5.00', '2.00', '1.00', '1.00', '1.00', '2.00', '']),
    ]

datasets = [
    (
        pandas.DataFrame(list(zip(itertools.cycle('AB'), range(10))), columns=['category', 'value']),
        pandas.Series(['0.00', '1.00', '2.00', '3.00', '4.00', '5.00', '6.00', '7.00', '8.00', '9.00']),
    ),(
        pandas.DataFrame(list(zip(itertools.cycle('AB'), [1, 1e6, 5, 5e9, 3, 2e8, 4, 4e6])), columns=['category', 'value']).groupby('category'),
        pandas.Series(['1.00', '1M', '5.00', '5,000M', '3.00', '200M', '4.00', '4M']),
    )

]

class TestCase(unittest.TestCase):
    def test_notation(self):
        for provided, expected in cases:
            self.assertEqual(notation.business(provided), expected)

    def test_missing(self):
        for provided, expected in missing_cases:
            self.assertEqual(notation.business(provided), expected)

    def test_upcast(self):
        for provided, expected in datasets:
            actual = provided.value.apply(notation.business)
        self.assertEqual(list(actual.index), list(expected.index))
        self.assertEqual(list(actual.values), list(expected.values))



# encoding: utf-8

import decimal
from math import copysign, floor, log

from .statistics import median
from .utils import ceil as _ceil
from .utils import floor as _floor
from .utils import bound, isnan, repel, replace, unwrap, vectorize


def order(value, base=10):
    if isnan(value):
        return None
    elif value == 0:
        return 0
    else:
        power = log(repel(value), base)
        return int(floor(power))


def e(exponent):
    value = str(abs(exponent))
    if exponent < 0:
        return 'E-' + value
    elif exponent > 0:
        return 'E+' + value
    else:
        return ''


def upcast(new, old):
    try:
        import pandas
        if isinstance(old, pandas.Series):
            return pandas.Series(new, index=old.index)
        else:
            return new
    except ImportError:
        return new


def extent(values, precision=3, statistic=median):
    l = min(values)
    r = max(values)
    exponent = order(statistic(values))
    digits = precision - exponent
    return (_floor(l, digits), _ceil(r, digits))


# TODO: just have this be an array and then zip this with
# range(-len(SI)//2*3, len(SI)//2*3, 3)
SI = {
     24: 'Y',
     21: 'Z',
     18: 'E',
     15: 'P',
     12: 'T',
      9: 'G',
      6: 'M',
      3: 'K',
      0: '',
     -3: 'm',
     -6: 'µ',
     -9: 'n',
    -12: 'p',
    -15: 'f',
    -18: 'a',
    -21: 'z',
    -24: 'y',
}


@unwrap
@vectorize
def human(value, digits=2):
    return '{:,}'.format(round(value, digits))


@unwrap
@vectorize
def scientific(value, precision=3):
    display = decimal.Context(prec=precision)
    value = decimal.Decimal(value).normalize(context=display)
    return display.to_sci_string(value)


@unwrap
@vectorize
def engineering(value, precision=3, prefix=False, prefixes=SI):
    """ Convert a number to engineering notation. """

    display = decimal.Context(prec=precision)
    value = decimal.Decimal(value).normalize(context=display)
    string = value.to_eng_string()

    if prefix:
        prefixes = {e(exponent): prefix for exponent, prefix in prefixes.items()}
        return replace(string, prefixes)
    else:
        return string


@unwrap
def business(values, precision=3, prefix=True, prefixes=SI, statistic=median, default=''):
    """
    Convert a list of numbers to the engineering notation appropriate to a
    reference point like the minimum, the median or the mean --
    think of it as "business notation".

    Any number will have at most the amount of significant digits of the
    reference point, that is, the function will round beyond the
    decimal point.

    For example, if the reference is `233K`, this function will turn
    1,175,125 into `1180K` and 11,234 into `11K` (instead of 1175K and
    11.2K respectively.) This can help enormously with readability.

    If the reference point is equal to or larger than E15 or
    equal to or smaller than E-15, E12 and E-12 become the
    reference point instead. (Petas and femtos are too
    unfamiliar to people to be easily comprehended.)
    """

    reference = statistic(values)

    if not reference:
        return upcast([''] * len(values), values)

    exponent = order(reference)
    e = bound(exponent - exponent % 3, -12, 12)
    # the amount of decimals is the precision minus the amount of digits
    # before the decimal point, which is one more than the relative order
    # of magnitude (for example, 10^5 can be represented as 100K, with
    # those three digits representing place values of 10^3, 10^4 and 10^5)
    d = precision - (1 + exponent - e)

    prefix = prefixes[e]

    strings = []
    for value in values:
        if isnan(value):
            strings.append('')
        else:
            normalized = value / 10.0 ** e
            # use `round` for rounding (beyond the decimal point if necessary)
            # use string formatting for padding to the right amount of decimals
            # and to hide decimals when necessary (by default, floats are always
            # displayed with a single decimal place, to distinguish them from
            # integers)
            relative_order = order(value) - exponent
            places = min(d - relative_order, d)
            normalized = round(normalized, places)
            strings.append('{0:,.{1}f}'.format(normalized, d) + prefix)

    return upcast(strings, values)



from . import notation, shortcuts, statistics, tests
from .notation import order, extent
from .notation import scientific, engineering, business
from .utils import quantize, floor, ceil, round

ballpark = business



from setuptools import find_packages, setup

setup(name='ballpark-fixed',
    description='Better human-readable numbers. Forked and fixed from Stijn',
    long_description=open('README.rst').read(),
    author='Endlessz',
    author_email='lckrep@protonmail.com',
    url='https://github.com/Endlesszombiez/python-ballpark',
    version='1.4.1',
    license='ISC',
    packages=find_packages(),
    keywords='human numbers format notation scientific engineering',
    install_requires=[],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Utilities',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.10'
        ],
    )
